#include "../aiger/aiger.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

const char * usage =
"usage: aigtopoly [ <option> ... ] <aiger>\n"
"\n"
"  -h        print this command line option summary and exit\n"
"  -i <str>  input prefix (default 'i')\n"
"  -o <str>  output prefix (default 's')\n"
"  -a <str>  AND gate prefix (default 'l')\n"
"\n"
"Encodes an AIGER file into polynomials.\n"
;

static aiger * model;

static void die (const char * msg, ...) {
  va_list ap;
  fputs ("*** aigtopoly: ", stderr);
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (1);
}

static char ** names;

static void print_field_polynomial (const char * name) {
  printf ("-%s+%s^2;\n", name, name);
}

int main (int argc, char ** argv) {
  const char * i = "i", * o = "s", * a = "l", * path = 0;
  for (int j = 1; j < argc; j++)
    if (!strcmp (argv[j], "-h")) fputs (usage, stdout), exit (0);
    else if (!strcmp (argv[j], "-i")) {
      if (++j == argc) die ("argument to '-i' missing");
      i = argv[j];
    } else if (!strcmp (argv[j], "-o")) {
      if (++j == argc) die ("argument to '-o' missing");
      o = argv[j];
    } else if (!strcmp (argv[j], "-a")) {
      if (++j == argc) die ("argument to '-a' missing");
      a = argv[j];
    } else if (argv[j][0] == '-')
      die ("invalid option '%s' (try '-h')", argv[j]);
    else if (path) die ("too many arguments '%d' and '%s'", path, argv[j]);
    else path = argv[j];
  if (!path) die ("AIGER model missing");
  model = aiger_init ();
  const char * err = aiger_open_and_read_from_file (model, path);
  if (err) die ("parse error: %s", err);
  if (model->num_latches) die ("can not handle latches");
  if (model->num_bad) die ("can not handle bad state properties");
  if (model->num_constraints) die ("can not handle constraints");
  if (model->num_justice) die ("can not handle justice properties");
  if (model->num_fairness) die ("can not handle fairness constraints");
  aiger_reencode (model);
  names = calloc (model->maxvar + 1, sizeof *names);
  int len = strlen (o);
  for (unsigned j = 0; j < model->num_inputs; j++) {
    char buffer[len + 20];
    sprintf (buffer, "%s%u", i, j);
    assert (model->inputs[j].lit/2 == j + 1);
    names[j + 1] = strdup (buffer);
    print_field_polynomial (buffer);
  }
  len = strlen (a);
  assert (model->maxvar - model->num_inputs == model->num_ands);
  for (unsigned j = model->num_inputs + 1; j <= model->maxvar; j++) {
    char buffer[len + 20];
    // sprintf (buffer, "%s%u", a, j);
    sprintf (buffer, "%s%u", a, 2*j);	// easier for debugging
    names[j] = strdup (buffer);
    aiger_and * and = aiger_is_and (model, 2*j);
    assert (and);
    printf ("-%s+", names[j]);
    unsigned r0 = and->rhs0;
    unsigned r1 = and->rhs1;
    const char * n0 = names[r0/2];
    const char * n1 = names[r1/2];
    if (!aiger_sign (r0) && !aiger_sign (r1)) {
      if (r0 == aiger_false || r1 == aiger_false) printf ("0");
      else if (r0 == r1) printf ("%s", n0);
      else printf ("%s*%s", n0, n1);
    } else if (!aiger_sign (r0) && aiger_sign (r1)) {
      if (r0 == aiger_false || r0 == aiger_not (r1)) printf ("0");
      else if (r1 == aiger_true) printf ("%s", n0);
      else printf ("%s-%s*%s", n0, n0, n1);
    } else if (aiger_sign (r0) && !aiger_sign (r1)) {
      if (r1 == aiger_false || r0 == aiger_not (r1)) printf ("0");
      else if (r0 == aiger_true) printf ("%s", n1);
      else printf ("%s-%s*%s", n1, n1, n0);
    } else {
      assert (aiger_sign (r0)), assert (aiger_sign (r1));
      if (r1 == aiger_true) printf ("%s", n0);
      else if (r0 == aiger_true) printf ("%s", n1);
      else printf ("1-%s-%s+%s*%s", n0, n1, n0, n1);
    }
    printf (";\n");
    print_field_polynomial (names[j]);
  }
  len = strlen (o);
  for (unsigned j = 0; j < model->num_outputs; j++) {
    char buffer[len + 20];
    sprintf (buffer, "%s%u", o, j);
    printf ("-%s+", buffer);
    aiger_symbol * symbol = model->outputs + j;
    unsigned lit = symbol->lit;
    if (lit == aiger_true) printf ("1");
    else if (lit == aiger_false) printf ("0");
    else {
      if (aiger_sign (lit)) printf ("1-");
      printf ("%s", names[lit/2]);
    }
    printf (";\n");
    print_field_polynomial (buffer);
  }
  aiger_reset (model);
  return 0;
}
