#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

const char * usage =
"usage: paceqspec [ <option> ... ] <bit-width>\n"
"\n"
"  -h       print this command line option summary and exit\n"
"  -1 <str> first outputs prefix (default 's')\n"
"  -2 <str> second output prefix (default 't')\n"
"  -e       print powers-of-two coefficients, e.g., '2^10' instead of '1024'\n"
"\n"
"\n"
"The <bit-width> should be a positive number 'n'.  The purpose of this\n"
"program is to generate a polynomial 'R' which is inconsistent with\n"
"two sets of polynomials describing gate-level circuits if and only if\n"
"those two circuits with outputs 's<0> ... s<n-1>' and 't<0> ... t<n-1>'\n"
"are logically equivalent.\n"
"\n"
"More precisely, let\n"
"\n"
"    n-1                n-1\n"
"A = sum 2^j s<j>   B = sum 2^j j<j>\n"
"    j=0                j=0\n"
"\n"
"then consider the following specification polynomial\n"
"\n"
"P = A - B\n"
"\n"
"which in turn is turned into 'R'\n"
"\n"
"R = 1 - y*P\n"
"\n"
"using the Rabinowitsch trick with a new variable 'y'.\n"
;

static void die (const char * msg, ...) {
  va_list ap;
  fputs ("*** paceqspec: ", stderr);
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (1);
}

static int use_exponent;
static char ** powers;

static char * times2 (const char * a) {
  int len = strlen (a), carry = 0;
  const char * p = a + len;
  char * res = malloc (len + 2) + len + 1;
  *res = 0;
  while (p > a) {
    int digit = *--p - '0';
    int tmp = 2*digit + carry;
    *--res = '0' + (tmp % 10);
    carry = (tmp > 9);
  }
  if (carry) *--res = '1';
  return res;
}

static void power (int i) {
  if (!i) return;
  if (use_exponent) printf ("2^%d*", i);
  else printf ("%s*", powers[i]);
}

int main (int argc, char ** argv) {
  const char * o1 = "s", * o2 = "t";
  int n = 0;

  for (int j = 1; j < argc; j++)
    if (!strcmp (argv[j], "-h")) fputs (usage, stdout), exit (0);
    else if (!strcmp (argv[j], "-1")) {
      if (++j == argc) die ("argument to '-1' missing");
      o1 = argv[j];
    } else if (!strcmp (argv[j], "-2")) {
      if (++j == argc) die ("argument to '-2' missing");
      o2 = argv[j];
    } else if (!strcmp (argv[j], "-e")) use_exponent = 1;
    else if (argv[j][0] == '-')
      die ("invalid option '%s' (try '-h')", argv[j]);
    else if (n) die ("too many arguments '%d' and '%s'", n, argv[j]);
    else if ((n = atoi (argv[1])) < 1)
      die ("invalid bit-width argument '%s' (try '-h')", argv[1]);
  if (!n) die ("bit-width argument missing (try '-h')");

  if (!use_exponent) {
    powers = malloc (2*n*sizeof *powers);
    powers[0] = strdup ("1");
    for (int j = 1; j < 2*n; j++) powers[j] = times2 (powers[j-1]);
  }

  printf ("1");

  for (int j = 0; j < n; j++)
    printf ("-"), power (j), printf ("%s%d*y", o1, j);
  for (int j = 0; j < n; j++)
    printf ("+"), power (j), printf ("%s%d*y", o2, j);

  printf (";\n");

  return 0;
}
