#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

const char * usage =
"usage: pacmultspec [ <option> ... ] <bit-width>\n"
"\n"
"  -h        print this command line option summary and exit\n"
"  -o <str>  output prefix (default 's')\n"
"  -i <str>  input prefix (default 'i')\n"
"  -e        print powers-of-two coefficients, e.g., '2^10' instead of '1024'\n"
"\n"
"Otherwise a polynomial 'R' is printed which is inconsistent with a\n"
"gate-level circuit implementing a multiplier of size 'n' where\n"
"'n' is determined by the given (positive) '<bit-width>' argument.\n"
"\n"
"The inputs are supposed to range from 'i0' to 'i<2n-1>' and the outputs from\n"
"'o<0>' to 'o<2n-1>', with 'i0' to 'i<n-1>' denoting the first operand 'A'\n"
"and 'i<n>' to 'i<2n-1>' the second operand 'B'.\n"
"\n"
"More precisely, the encoding is as follows\n"
"\n"
"    n-1                n-1                 2n-1\n"
"A = sum 2^j i<j>   B = sum 2^j i<j+n>   C = sum 2^j o<j>\n"
"    j=0                j=0                  j=0\n"
"\n"
"and we get the following specification polynomial\n"
"\n"
"P = C - A * B\n"
"\n"
"which in turn is turned into 'R'\n"
"\n"
"R = 1 - y*P\n"
"\n"
"using the Rabinowitsch trick with a new variable 'y'.\n"
;

static void die (const char * msg, ...) {
  va_list ap;
  fputs ("*** pacmultspec: ", stderr);
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (1);
}

static int use_exponent;
static char ** powers;

static char * times2 (const char * a) {
  int len = strlen (a), carry = 0;
  const char * p = a + len;
  char * res = malloc (len + 2) + len + 1;
  *res = 0;
  while (p > a) {
    int digit = *--p - '0';
    int tmp = 2*digit + carry;
    *--res = '0' + (tmp % 10);
    carry = (tmp > 9);
  }
  if (carry) *--res = '1';
  return res;
}

static void power (int i) {
  if (!i) return;
  if (use_exponent) printf ("2^%d*", i);
  else printf ("%s*", powers[i]);
}

int main (int argc, char ** argv) {
  const char * i = "i", * o = "s";
  int n = 0;

  for (int j = 1; j < argc; j++)
    if (!strcmp (argv[j], "-h")) fputs (usage, stdout), exit (0);
    else if (!strcmp (argv[j], "-i")) {
      if (++j == argc) die ("argument to '-i' missing");
      i = argv[j];
    } else if (!strcmp (argv[j], "-o")) {
      if (++j == argc) die ("argument to '-o' missing");
      o = argv[j];
    } else if (!strcmp (argv[j], "-e")) use_exponent = 1;
    else if (argv[j][0] == '-')
      die ("invalid option '%s' (try '-h')", argv[j]);
    else if (n) die ("too many arguments '%d' and '%s'", n, argv[j]);
    else if ((n = atoi (argv[1])) < 1)
      die ("invalid bit-width argument '%s' (try '-h')", argv[1]);
  if (!n) die ("bit-width argument missing (try '-h')");

  if (!use_exponent) {
    powers = malloc (2*n*sizeof *powers);
    powers[0] = strdup ("1");
    for (int j = 1; j < 2*n; j++) powers[j] = times2 (powers[j-1]);
  }

  printf ("1");

  for (int j = 0; j < n; j++)
    for (int k = 0; k < n; k++)
      printf ("-"), power (j + k), printf ("%s%d*%s%d*y", i, j, i, n+k);

  for (int j = 0; j < 2*n; j++)
    printf ("+"), power (j), printf ("%s%d*y", o, j);

  printf (";\n");

  return 0;
}
