/*------------------------------------------------------------------------*/
/* Copyright (C) 2018, Armin Biere, Johannes Kepler University Linz */
/*------------------------------------------------------------------------*/

#define VERSION "000"

static const char * usage =
"pactrim [ <option> ... ]  [ (<polynomials> <proof>) | <mixed> ]\n"
"\n"
"where <option> is one of the following\n"
"\n"
"  -h             print this command line option summary and exit\n"
"\n"             
"  -S             only check inferences but not that proof is a refutation\n"
"                 (so do not check that a non-zero constant is inferred)\n"
"\n"             
"  -c <core>      write core of original polynomials to this file\n"
"\n"             
"  -l <lemmas>    write core of proof lemmas to this file\n"
"\n"
"  -u <unused>    write unused original polynomials to this file\n"
"\n"
"  -r <redundant> write redundant proof lemmas to this file\n"
"\n"
"  --print        print polynomials in failing checks for better diagnosis\n"
"\n"
"  --sparse       use spaces to make printed polynomials easier to read\n"
"\n"
"The <polynomials> argument should point to a file with the\n"
"original set of polynomials and <proof> is a path to a proof file\n"
"interpreted as a sequence of inferences in the polynomial calculus.\n"
"The tool checks that all inferences in the sequence are correct.\n"
"If only one file is specified the tool goes into mixed mode and\n"
"assumes that both original polynomial and inferences are in the file.\n"
"This also applies if no file is specified and mixed is read from 'stdin'.\n" 
"If '-S' is not specified, it is further checked that the proof\n"
"is a refutation, i.e., contains a non-zero constant polynomial.\n"
"The exit code is zero if and only if all checks succeed.\n"
;

/*------------------------------------------------------------------------*/

#include <assert.h>
#include <ctype.h>
#include <inttypes.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*------------------------------------------------------------------------*/

// error and verbose messages

static void die (const char * msg, ...) {
  fflush (stdout);
  fprintf (stderr, "*** 'pactrim' error: ");
  va_list ap;
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fprintf (stderr, "\n");
  fflush (stderr);
  exit (1);
}

static void msg (const char * msg, ...) {
  fprintf (stdout, "[pactrim] ");
  va_list ap;
  va_start (ap, msg);
  vfprintf (stdout, msg, ap);
  va_end (ap);
  fprintf (stdout, "\n");
  fflush (stdout);
}

/*------------------------------------------------------------------------*/

// memory management routines maintaining a count of allocated memory

static size_t current_allocated_bytes;
static size_t max_allocated_bytes;

static void inc_allocated_bytes (size_t bytes) {
  current_allocated_bytes += bytes;
  if (current_allocated_bytes > max_allocated_bytes)
    max_allocated_bytes = current_allocated_bytes;
}

static void dec_allocated_bytes (size_t bytes) {
  assert (current_allocated_bytes >= bytes);
  current_allocated_bytes -= bytes;
}

static void * allocate (size_t bytes) {
  if (!bytes) return 0;
  void * res = malloc (bytes);
  if (bytes) {
    if (res) memset (res, 0, bytes);
    else die ("out-of-memory allocating %zu bytes", bytes);
  }
  inc_allocated_bytes (bytes);
  return res;
}

static void deallocate (void * ptr, size_t bytes) {
  dec_allocated_bytes (bytes);
  free (ptr);
}

static void * reallocate (void * ptr, size_t old_bytes, size_t new_bytes) {
  dec_allocated_bytes (old_bytes);
  void * res = realloc (ptr, new_bytes);
  if (res) {
    if (new_bytes > old_bytes)
      memset (res + old_bytes, 0, new_bytes - old_bytes);
  } else die ("out-of-memory reallocating %zu bytes", new_bytes);
  inc_allocated_bytes (new_bytes);
  return res;
}

/*------------------------------------------------------------------------*/

// 'polymorphic' array management routine maintain a count of allocated memory

#define ALLOCATE(P,N) \
do { \
  size_t BYTES = (N) * sizeof *(P); \
  (P) = allocate (BYTES); \
} while (0)

#define DEALLOCATE(P,N) \
do { \
  size_t BYTES = (N) * sizeof *(P); \
  deallocate ((P), BYTES); \
} while (0)

#define REALLOCATE(P,O,N) \
do { \
  size_t OLD_BYTES = (O) * sizeof *(P); \
  size_t NEW_BYTES = (N) * sizeof *(P); \
  (P) = reallocate ((P), OLD_BYTES, NEW_BYTES); \
} while (0)

/*------------------------------------------------------------------------*/

// some smaller helper macros

#define SWAP(TYPE,A,B) \
do { \
  TYPE TMP = (A); \
  (A) = (B); \
  (B) = TMP; \
} while (0)

#define MAX(A,B) \
  ((A) < (B) ? (B) : (A))

/*------------------------------------------------------------------------*/

// Random odd 64 numbers for constructing hash functions.

static uint64_t nonces[32];
static const size_t num_nonces = sizeof nonces / sizeof nonces[0];

static uint16_t rand16 () {
  int tmp = rand ();
  assert (tmp >= 0);
  uint16_t res = tmp ^ (tmp >> 16);
  return res;
}

static uint64_t rand64 () {
  uint64_t res = 0;
  for (unsigned i = 0; i < 64; i += 16)
    res |= rand16 () << i;
  return res;
}

static void init_nonces () {
  srand (42);
  for (size_t i = 0; i < num_nonces; i++)
    nonces[i] = rand64 () | 1;
}

/*------------------------------------------------------------------------*/

// String functions.

static char * duplicate_string (const char * str) {
  if (!str) return 0;
  size_t bytes = strlen (str) + 1;
  char * res = allocate (bytes);
  strcpy (res, str);
  return res;
}

static void deallocate_string (char * str) {
  deallocate (str, strlen (str) + 1);
}

static uint64_t hash_string (const char * str) {
  uint64_t res = 0;
  size_t i = 0;
  for (const char * p = str; *p; p++) {
    res += *p;
    res *= nonces[i++];
    if (i == num_nonces) i = 0;
  }
  return res;
}

/*------------------------------------------------------------------------*/

// Variables are kept canonical using a global hash table similar to a
// symbol table in an ordinary compiler.

typedef struct Variable Variable;

struct Variable {
  char * name;			// variable name
  uint64_t hash;		// saved hash value
  Variable * next;		// hash collision chain link
};

static uint64_t num_variables;
static uint64_t size_variables;
static Variable ** variable_table;

static void enlarge_variables () {
  uint64_t new_size_variables = size_variables ? 2*size_variables : 1;
  Variable ** new_variable_table;
  ALLOCATE (new_variable_table, new_size_variables);
  for (uint64_t i = 0; i < size_variables; i++) {
    for (Variable * v = variable_table[i], * n; v; v = n) {
      uint64_t h = v->hash & (new_size_variables - 1);
      n = v->next;
      v->next = new_variable_table[h];
      new_variable_table[h] = v;
    }
  }
  DEALLOCATE (variable_table, size_variables);
  variable_table = new_variable_table;
  size_variables = new_size_variables;
}

static int is_valid_variable_letter (char ch) {
  if (ch == '_') return 1;
  if ('0' <= ch && ch <= '9') return 1;
  if ('a' <= ch && ch <= 'z') return 1;
  if ('A' <= ch && ch <= 'Z') return 1;
  return 0;
}

static int is_valid_variable_first_letter (char ch) {
  if (ch == '_') return 1;
  if ('a' <= ch && ch <= 'z') return 1;
  if ('A' <= ch && ch <= 'Z') return 1;
  return 0;
}

#ifndef NDEBUG

static int is_valid_variable_name (const char * name) {
  if (!name) return 0;
  if (!is_valid_variable_first_letter (name[0])) return 0;
  for (const char * p = name + 1; *p; p++)
    if (!is_valid_variable_letter (*p)) return 0;
  return 1;
}

#endif

static Variable * new_variable (const char * name) {
  assert (is_valid_variable_name (name));
  if (num_variables == size_variables) enlarge_variables ();
  const uint64_t hash = hash_string (name);
  const uint64_t h = hash & (size_variables - 1);
  Variable ** p, * res;
  for (p = variable_table + h;
       (res = *p) && strcmp (res->name, name);
       p = &res->next)
    ;
  if (!res) {
    *p = res = allocate (sizeof *res);
    res->name = duplicate_string (name);
    res->hash = hash;
    res->next = 0;
    num_variables++;
  }
  return res;
}

static void deallocate_variables () {
  for (uint64_t i = 0; i < size_variables; i++) {
    for (Variable * v = variable_table[i], * n; v; v = n) {
      n = v->next;
      deallocate_string (v->name);
      deallocate (v, sizeof *v);
    }
  }
  DEALLOCATE (variable_table, size_variables);
}

static void print_variable (Variable * v, FILE * file) {
  fputs (v->name, file);
}

static int cmp_variable (Variable * a, Variable * b) {
  if (a == b) return 0;
  return strcmp (a->name, b->name);
}

/*------------------------------------------------------------------------*/

// Arbitrary precision number arithmetic routines for integers.

// We use a string based simple ASCII encoding for numbers.  This is of
// course slower than a binary word list encoding and much slower than using
// a standard library such as GNU MP for this purpose.  First we do not
// think that these routines are a bottle neck in our case.  Further
// they are simpler, maybe even verifiable, and easy to debug.  The main
// reasons is to be able to compute hash values for numbers without
// resorting to the internals of the GNU MP library, which does not support
// hashing.  Moreover, it keeps library dependency of this code simpler.

// We do not a have canonical representation of numbers assuming that
// most coefficients are different.  We might as future extension add some
// fly weight global constants, say for '1', '0', -1', etc.

typedef char Number;

static uint64_t max_numbers;
static uint64_t current_numbers;
static uint64_t total_numbers;

static int sign_of_number (const char * a) { return *a == '-' ? -1 : 1; }

#ifndef NDEBUG

static int normalized_natural (const Number * a) {
  if (!*a) return 0;
  if (sign_of_number (a) < 0) return 0;
  if (a[0] == '0' && a[1]) return 0;
  int ch;
  while ((ch = *a++))
    if (ch < '0' || ch > '9') return 0;
  return 1;
}

static int normalized_number (const Number * a) {
  if (!*a) return 0;
  if (sign_of_number (a) < 0 && *++a == '0') return 0;
  return normalized_natural (a);
}

#endif

static int is_zero_number (const Number * a) {
  assert (normalized_number (a));
  return a[0] == '0' && !a[1];
}

static int is_one_number (const Number * a) {
  assert (normalized_number (a));
  return a[0] == '1' && !a[1];
}

static int is_minus_one_number (const Number * a) {
  assert (normalized_number (a));
  return a[0] == '-' && a[1] == '1' && !a[2];
}

#ifndef NDEBUG

static int is_small_number (const Number * a) {
  if (is_zero_number (a)) return 1;
  if (is_one_number (a)) return 1;
  if (is_minus_one_number (a)) return 1;
  return 0;
}

#endif

static void inc_numbers () {
  total_numbers++;
  current_numbers++;
  if (current_numbers > max_numbers) max_numbers = current_numbers;
}

static Number * new_number (const char * str) {
  const char * normalized = str;
  int sign = sign_of_number (str);
  if (sign < 0) normalized++;
  while (*normalized == '0') normalized++;
  if (!*normalized) normalized = "0", sign = 1;
  size_t nlen = strlen (normalized);
  Number * res;
  if (sign < 0) {
    res = allocate (nlen + 2);
    if (res) {
      strcpy (res + 1, normalized);
      *res = '-';
    }
  } else res = duplicate_string (normalized);
  inc_numbers ();
  assert (normalized_number (res));
  return res;
}

static Number * copy_number (const Number * a) {
  assert (normalized_number (a));
  inc_numbers ();
  return duplicate_string (a);
}

static void delete_number (Number * n) {
  assert (current_numbers > 0);
  current_numbers--;
  deallocate_string (n);
}

static uint64_t hash_number (const Number * n) {
  assert (normalized_number (n));
  return hash_string (n);
}

static int cmp_numbers (const Number * a, const Number * b) {
  const int asign = sign_of_number (a);
  const int bsign = sign_of_number (b);
  if (asign < bsign) return -1;
  if (asign > bsign) return 1;
  const int sign = asign;
  assert (normalized_number (a));
  assert (normalized_number (b));
  const size_t alen = strlen (a);
  const size_t blen = strlen (b);
  if (alen < blen) return -sign;
  if (alen > blen) return sign;
  while (*a && *a == *b)
    a++, b++;
  if (*a < *b) return -sign;
  if (*a > *b) return sign;
  return 0;
}

// Sub section for UNSIGNED arbitrary precision number arithmetic.  They all
// have '... natural ...' in their name and are used by the actual signed
// routines in the next section after removing the sign.

static Number *
add_naturals (int sign, const Number * a, const Number * b) {
  assert (normalized_natural (a));
  assert (normalized_natural (b));
  const size_t alen = strlen (a);
  const size_t blen = strlen (b);
  const size_t maxlen = MAX (alen, blen);
  const char * p = a + alen;
  const char * q = b + blen;
  char accu[maxlen + 3];
  char * r = accu + sizeof accu;
  int carry = 0;
  *--r = 0;
  for (size_t i = 0; i < maxlen; i++) {
    const int achar = a < p ? *--p : '0';
    const int bchar = b < q ? *--q : '0';
    int tmp = achar + bchar + carry - '0';
    if (tmp > '9') tmp -= 10, carry = 1;
    else carry = 0;
    *--r = tmp;
  }
  if (carry) *--r = '1';
  if (sign < 0) *--r = '-';
  return new_number (r);
}

static Number *
subtract_naturals (int sign, const Number * a, const Number * b) {
  assert (normalized_natural (a));
  assert (normalized_natural (b));
  assert (cmp_numbers (a, b) >= 0);
  const size_t alen = strlen (a);
  const size_t blen = strlen (b);
  const char * p = a + alen;
  const char * q = b + blen;
  char accu[alen + 2];
  char * r = accu + sizeof accu;
  int carry = 0;
  *--r = 0;
  for (size_t i = 0; i < alen; i++) {
    const int achar = a < p ? *--p : '0';
    const int bchar = b < q ? *--q : '0';
    int tmp = achar - bchar - carry + '0';
    if (tmp < '0') tmp += 10, carry = 1;
    else carry = 0;
    *--r = tmp;
  }
  assert (!carry);
  if (sign < 0) *--r = '-';
  return new_number (r);
}

static Number *
multiply_natural_step (const Number * a,
                       const char digit, const size_t exponent) {
  assert (normalized_natural (a));
  if (digit == '0') return copy_number ("0");
  if (!exponent && digit == '1') return copy_number (a);
  const size_t alen = strlen (a);
  const char * p = a + alen;
  char accu[alen + exponent + 2];
  char * r = accu + sizeof accu;
  int carry = 0;
  *--r = 0;
  for (size_t i = 0; i < exponent; i++)
    *--r = '0';
  for (size_t i = 0; i < alen; i++) {
    const int achar = a < p ? *--p : '0';
    int tmp = (achar - '0') * (digit - '0') + carry;
    carry = tmp / 10;
    *--r = (tmp % 10) + '0';
  }
  if (carry) *--r = carry + '0';
  return new_number (r);
}

// Back to signed arbitrary precision number arithmetic routines.

static Number * add_numbers (const Number * a, const Number * b) {
  assert (normalized_number (a));
  assert (normalized_number (b));
  int asign = sign_of_number (a);
  int bsign = sign_of_number (b);
  if (asign < 0) a++;
  if (bsign < 0) b++;
  assert (normalized_natural (a));
  assert (normalized_natural (b));
  if (asign == bsign) return add_naturals (asign, a, b);
  const int cmp = cmp_numbers (a, b);
  if (!cmp) return copy_number ("0");
  if (cmp < 0) {
    SWAP (const Number *, a, b);
    SWAP (const int, asign, bsign);
  }
  return subtract_naturals (asign, a, b);
}

static Number * negate_number (const Number * a) {
  assert (normalized_number (a));
  if (sign_of_number (a) < 0) return copy_number (a + 1);
  if (is_zero_number (a)) return copy_number ("0");
  const size_t alen = strlen (a);
  Number * res = allocate (alen + 2);
  if (res) {
    strcpy (res + 1, a);
    *res = '-';
  }
  inc_numbers ();
  return res;
}

static Number * multiply_numbers (const Number * a, const Number * b) {
  assert (normalized_number (a));
  assert (normalized_number (b));
  if (is_one_number (a)) return copy_number (b);
  if (is_zero_number (a)) return copy_number ("0");
  if (is_minus_one_number (a)) return negate_number (b);
  if (is_one_number (b)) return copy_number (a);
  if (is_zero_number (b)) return copy_number ("0");
  if (is_minus_one_number (b)) return negate_number (a);
  assert (!is_small_number (a));
  assert (!is_small_number (b));
  int asign = sign_of_number (a);
  int bsign = sign_of_number (b);
  const int sign = asign * bsign;
  if (asign < 0) a++;
  if (bsign < 0) b++;
  assert (normalized_natural (a));
  assert (normalized_natural (b));
  size_t alen = strlen (a);
  size_t blen = strlen (b);
  if (alen > blen) {
    SWAP (const Number *, a, b);
    SWAP (int, asign, bsign);               // redundant
    SWAP (int, alen, blen);
  }
  Number * res = copy_number ("0");
  for (size_t i = 0; i < alen; i++) {
    Number * step = multiply_natural_step (b, a[alen - i - 1], i);
    Number * tmp = add_naturals (1, res, step);
    delete_number (step);
    delete_number (res);
    res = tmp;
  }
  if (sign < 0) {
    Number * tmp = negate_number (res);
    delete_number (res);
    res = tmp;
  }
  return res;
}

static void print_number (Number * c, FILE * file) { fputs (c, file); }

static void print_abs_number (Number * c, FILE * file) {
  fputs (*c == '-' ? c + 1 : c, file);
}

/*------------------------------------------------------------------------*/

// Routines to manipulate powers of variables.

typedef struct Power Power;

// These powers are just pairs of variables and exponents and since
// variables are never deleted and exponents are just machine words we do
// not need any heap manipulation here.

struct Power {
  Variable * variable;
  uint64_t exponent;
};

static Power new_power (Variable * variable, uint64_t exponent) {
  Power res = { variable, exponent };
  return res;
}

static void print_power (Power p, FILE * file) {
  print_variable (p.variable, file);
  if (p.exponent != 1) {
    fputc ('^', file);
    fprintf (file, "%" PRIu64, p.exponent);
  }
}

static int cmp_powers (const Power * a, const Power * b) {
  return cmp_variable (a->variable, b->variable);
}

/*------------------------------------------------------------------------*/

// We provide a global stack of powers for sorting.

static uint64_t num_powers;
static uint64_t size_powers;
static Power * powers;

static void enlarge_powers () {
  size_t new_size_powers = size_powers ? 2*size_powers : 1;
  REALLOCATE (powers, size_powers, new_size_powers);
  size_powers = new_size_powers;
}

static void push_power (Power p) {
  if (size_powers == num_powers) enlarge_powers ();
  powers[num_powers++] = p;
}

static void clear_powers () { num_powers = 0; }

static void deallocate_powers () { DEALLOCATE (powers, size_powers); }

// Sort powers according to the variable order.

static void sort_powers () {
  qsort (powers, num_powers, sizeof *powers,
    (int(*)(const void *, const void *)) cmp_powers);
}

// Merge powers on the stack and flush out those with zero exponent too.

static void merge_powers () {
  Power * a = 0;
  size_t i = 0;
  for (size_t j = 0; j < num_powers; j++) {
    Power * b = powers + j;
    if (!b->exponent) continue;
    if (a && a->variable == b->variable) {
      if (UINT64_MAX - a->exponent < b->exponent)
	die ("maximum exponent range exhausted while merging powers");
      a->exponent += b->exponent;
    } else {
      if (a) assert (cmp_variable (a->variable, b->variable) < 0);
      a = powers + i;
      *a = *b;
      i++;
    }
  }
  num_powers = i;
}

/*------------------------------------------------------------------------*/

// Monomials are sorted products of powers of variables.

// They are in principle stored as sorted singly linked lists (through the
// 'rest' pointer) but all lists and suffixes are stored in a hash table to
// make them canonical.  In contrast to variables, which are assumed to be
// live much longer, we support deletion of monomials through reference
// counting.  This should save memory and eventually will also allow to
// properly support deletion lines in proofs.

typedef struct Monomial Monomial;

struct Monomial {
  Variable * variable;
  uint64_t exponent;	// non-zero exponent
  Monomial * rest;	// singly linked list
  uint64_t degree;	// degree of rest + exponent
  uint64_t ref;		// reference counter
  uint64_t hash;	// saved hash value
  Monomial * next;	// collision chain
};

// Hash table to canonical store monomials.

static uint64_t size_monomials;
static uint64_t current_monomials;
static Monomial ** monomial_table;

static void enlarge_monomials () {
  uint64_t new_size_monomials = size_monomials ? 2*size_monomials : 1;
  Monomial ** new_monomial_table;
  ALLOCATE (new_monomial_table, new_size_monomials);
  for (uint64_t i = 0; i < size_monomials; i++) {
    for (Monomial * m = monomial_table[i], * n; m; m = n) {
      uint64_t h = m->hash & (new_size_monomials - 1);
      n = m->next;
      m->next = new_monomial_table[h];
      new_monomial_table[h] = m;
    }
  }
  DEALLOCATE (monomial_table, size_monomials);
  monomial_table = new_monomial_table;
  size_monomials = new_size_monomials;
}

static uint64_t
compute_hash_monomial (Variable * variable,
                       uint64_t exponent, Monomial * rest) {
  assert (variable);
  uint64_t res = rest ? rest->hash : 0;
  res *= nonces[0];
  res += exponent;
  res *= nonces[1];
  res += variable->hash;
  res *= nonces[2];
  return res;
}

static int
match_monomial (Monomial * m,
                Variable * variable, uint64_t exponent, Monomial * rest) {
  if (m->variable != variable) return 0;
  if (m->exponent != exponent) return 0;
  if (m->rest != rest) return 0;
  return 1;
}

static Monomial * copy_monomial (Monomial * m) {
  if (!m) return m;
  assert (m->ref > 0);
  m->ref++;
  assert (m->ref);
  return m;
}

// Statistic counters for monomials.

static uint64_t max_monomials;
static uint64_t total_monomials;
static uint64_t hits_monomials;
static uint64_t searched_monomials;
static uint64_t collisions_monomials;
static uint64_t max_degree;

// First low-level allocation and deallocation of monomials.

static Monomial *
allocate_monomial (Variable * variable,
                   uint64_t exponent, Monomial * rest) {
  Monomial * res = allocate (sizeof *res);
  res->variable = variable;
  res->exponent = exponent;
  res->rest = copy_monomial (rest);
  if (!rest) res->degree = exponent;
  else if (UINT64_MAX - exponent < rest->degree)
    die ("maximum degree exhausted");
  else res->degree = exponent + rest->degree;
  if (res->degree > max_degree) max_degree = res->degree;
  total_monomials++;
  current_monomials++;
  if (current_monomials > max_monomials) max_monomials = current_monomials;
  return res;
}

static void deallocate_monomial (Monomial * m) {
  assert (current_monomials);
  current_monomials--;
  deallocate (m, sizeof *m);
}

static void deallocate_monomials () {
  for (uint64_t i = 0; i < size_monomials; i++) {
    for (Monomial * m = monomial_table[i], *n; m; m = n) {
      n = m->next;
      deallocate_monomial (m);
    }
  }
  DEALLOCATE (monomial_table, size_monomials);
}

// Then searching and adding a monomial extended by one power.

static Monomial *
new_monomial (Variable * variable, uint64_t exponent, Monomial * rest) {
  if (!exponent) return copy_monomial (rest);
  if (rest) assert (cmp_variable (variable, rest->variable) < 0);
  if (current_monomials == size_monomials) enlarge_monomials ();
  const uint64_t hash = compute_hash_monomial (variable, exponent, rest);
  const uint64_t h = hash & (size_monomials - 1);
  searched_monomials++;
  Monomial * res;
  for (res = monomial_table[h];
       res && !match_monomial (res, variable, exponent, rest);
       res = res->next) {
    collisions_monomials++;
  }
  if (res) {
    hits_monomials++;
    res->ref++;
  } else {
    res = allocate_monomial (variable, exponent, rest);
    res->ref = 1;
    res->hash = hash;
    res->next = monomial_table[h];
    monomial_table[h] = res;
  }
  return res;
}

// Decrements the reference count of a monomial, and actually deletes a
// monomial if its reference count goes to zero.  In this case it also
// removes it from the hash table and applies the same procedure to the
// suffix 'rest'.

static void delete_monomial (Monomial * m) {
  while (m) {
    assert (m->ref > 0);
    if (--m->ref > 0) break;
    Monomial * rest = m->rest;
    const uint64_t h = m->hash & (size_monomials - 1);
    Monomial ** p = monomial_table + h, * n;
    while ((n = *p) != m) assert (n), p = &n->next;
    *p = n->next;
    deallocate_monomial (m);
    m = rest;
  }
}

static void print_monomial (Monomial * m, FILE * file) {
  if (!m) fputc ('1', file);
  else {
    Monomial * n = m;
    while (n) {
      if (n != m) fputc ('*', file);
      print_power (new_power (n->variable, n->exponent), file);
      n = n->rest;
    }
  }
}

static Monomial * multiply_monomials (Monomial * a, Monomial * b) {
  if (!a) return copy_monomial (b);
  if (!b) return copy_monomial (a);
  const int cmp = cmp_variable (a->variable, b->variable);
  Monomial * arest, * brest;
  Variable * variable;
  uint64_t exponent;
  if (cmp < 0) {
    arest = a->rest;
    brest = b;
    exponent = a->exponent;
    variable = a->variable;
  } else if (cmp > 0) {
    arest = a;
    brest = b->rest;
    exponent = b->exponent;
    variable = b->variable;
  } else {
    assert (a->variable == b->variable);
    arest = a->rest;
    brest = b->rest;
    if (UINT64_MAX - a->exponent < b->exponent)
      die ("maximum exponent range exhausted while multiplying monomials");
    exponent = a->exponent + b->exponent;
    variable = a->variable;
  }
  Monomial * rest = multiply_monomials (arest, brest);
  Monomial * res = new_monomial (variable, exponent, rest);
  delete_monomial (rest);
  return res;
}

static int cmp_monomials (const Monomial * a, const Monomial * b) {
  while (a || b) {
    if (a == b) return 0;
    if (!a && b) return -1;
    if (a && !b) return 1;
    const int res = cmp_variable (a->variable, b->variable);
    if (res) return res;
    if (a->exponent > b->exponent) return -1;
    if (a->exponent < b->exponent) return 1;
    a = a->rest;
    b = b->rest;
  }
  return 0;
}

static uint64_t hash_monomial (const Monomial * a) {
  return a ? a->hash : 0;
}

static uint64_t degree_monomial (const Monomial * a) {
  return a ? a->degree : 0;
}

// Building a monomial from sorted and merges powers.

static Monomial * build_monomial () {
  Monomial * res = 0;
  int i = num_powers;
  while (i > 0) {
    Power p = powers[--i];
    Monomial * tmp = new_monomial (p.variable, p.exponent, res);
    delete_monomial (res);
    res = tmp;
  }
  return res;
}

/*------------------------------------------------------------------------*/

// Terms are monomials with a coefficient (currently only integers).

typedef struct Term Term;

struct Term {
  Number * number;		// Local coefficient.
  Monomial * monomial;		// Canonical monomial.
  uint64_t ref;
};

static int sign_of_term (Term * a) { return sign_of_number (a->number); }

static int is_zero_term (Term * a) { return is_zero_number (a->number); }

static int is_non_zero_constant_term (Term * a) {
  if (a->monomial) return 0;
  return !is_zero_number (a->number);
}

static void set_term_monomial (Term * a, Monomial * m) {
  assert (normalized_number (a->number));
  if (is_zero_number (a->number)) a->monomial = 0;
  else a->monomial = copy_monomial (m);
}

static Term * allocate_term () {
  Term * res = allocate (sizeof *res);
  res->ref = 1;
  return res;
}

static Term * new_term (const Number * number, Monomial * monomial) {
  Term * res = allocate_term ();
  res->number = copy_number (number);
  set_term_monomial (res, monomial);
  return res;
}

static Term * copy_term (Term * a) {
  assert (a->ref > 0);
  a->ref++;
  return a;
}

static void delete_term (Term * a) {
  assert (a->ref > 0);
  if (--a->ref > 0) return;
  delete_number (a->number);
  delete_monomial (a->monomial);
  deallocate (a, sizeof *a);
}

static Term * add_terms (Term * a, Term * b) {
  if (is_zero_term (a)) return copy_term (b);
  if (is_zero_term (b)) return copy_term (a);
  assert (a->monomial == b->monomial);
  Term * res = allocate_term ();
  res->number = add_numbers (a->number, b->number);
  set_term_monomial (res, a->monomial);
  return res;
}

static Term * multiply_terms (Term * a, Term * b) {
  Term * res = allocate_term ();
  res->number = multiply_numbers (a->number, b->number);
  if (is_zero_number (res->number)) res->monomial = 0;
  else res->monomial = multiply_monomials (a->monomial, b->monomial);
  return res;
}

static void print_term (Term * a, FILE * file) {
  if (!a->monomial) print_number (a->number, file);
  else {
    assert (!is_zero_number (a->number));
    if (is_one_number (a->number))
      print_monomial (a->monomial, file);
    else if (is_minus_one_number (a->number))
      fputc ('-', file), print_monomial (a->monomial, file);
    else {
      print_number (a->number, file);
      fputc ('*', file);
      print_monomial (a->monomial, file);
    }
  }
}

static void print_abs_term (Term * a, FILE * file) {
  if (!a->monomial) print_abs_number (a->number, file);
  else {
    assert (!is_zero_number (a->number));
    if (is_one_number (a->number) ||
        is_minus_one_number (a->number))
      print_monomial (a->monomial, file);
    else {
      print_abs_number (a->number, file);
      fputc ('*', file);
      print_monomial (a->monomial, file);
    }
  }
}

static int cmp_terms (Term * a, Term * b) {
  const int res = cmp_monomials (a->monomial, b->monomial);
  if (res) return res;
  return cmp_numbers (a->number, b->number);
}

static uint64_t hash_term (Term * a) {
  uint64_t res = hash_number (a->number);
  res *= nonces[0];
  res += hash_monomial (a->monomial);
  res *= nonces[1];
  return res;
}

/*------------------------------------------------------------------------*/

// Stack to sort and merge terms.

static size_t num_terms;
static size_t size_terms;
static Term ** terms;

static void enlarge_terms () {
  size_t new_size_terms = size_terms ? 2*size_terms : 1;
  REALLOCATE (terms, size_terms, new_size_terms);
  size_terms = new_size_terms;
}

static void push_term (Term * t) {
  if (size_terms == num_terms) enlarge_terms ();
  terms[num_terms++] = copy_term (t);
}

static void clear_terms () {
  for (size_t i = 0; i < num_terms; i++)
    delete_term (terms[i]);
  num_terms = 0;
}

static void deallocate_terms () { DEALLOCATE (terms, size_terms); }

static int cmp_terms_for_qsort (const void * p, const void * q) {
  Term * a = *(Term **) p;
  Term * b = *(Term **) q;
  return cmp_terms (a, b);
}

// Sort terms according to monomial order.

static void sort_terms () {
  qsort (terms, num_terms, sizeof *terms, cmp_terms_for_qsort);
}

// Merge terms with the same monomials and flush zero terms.

static void merge_terms () {
  Term * a = 0;
  size_t i = 0;
  for (size_t j = 0; j < num_terms; j++) {
    Term * b = terms[j];
    if (is_zero_term (b)) delete_term (b);
    else if (a && a->monomial == b->monomial) {
      Term * c = add_terms (a, b);
      assert (i > 0);
      assert (terms[i-1] == a);
      delete_term (a);
      delete_term (b);
      terms[i-1] = c;
      a = c;
    } else {
      if (a) assert (cmp_terms (a, b) < 0);
      terms[i++] = b;
      a = b;
    }
  }
  num_terms = i;
}

static uint64_t degree_term (Term * term) {
  return degree_monomial (term->monomial);
}

/*------------------------------------------------------------------------*/

typedef struct Polynomial Polynomial;

struct Polynomial {
  Term * term;
  Polynomial * rest;
};

static int is_non_zero_constant_polynomial (Polynomial * p) {
  if (!p) return 0;
  return !p->rest && is_non_zero_constant_term (p->term);
}

static Polynomial * new_polynomial (Term * term,  Polynomial * rest) {
  if (is_zero_term (term)) return rest;
  if (rest) assert (cmp_terms (term, rest->term) < 0);
  Polynomial * res = allocate (sizeof *res);
  res->term = copy_term (term);
  res->rest = rest;
  return res;
}

static void delete_polynomial (Polynomial * a) {
  while (a) {
    Polynomial * rest = a->rest;
    delete_term (a->term);
    deallocate (a, sizeof *a);
    a = rest;
  }
}

static int sparse_polynomial_printing = 0;

static void print_polynomial (Polynomial * p, FILE * file) {
  if (!p) fputc ('0', file);
  else {
    print_term (p->term, file);
    for (Polynomial * q = p->rest; q; q = q->rest) {
      int sign = sign_of_term (q->term);
      if (sparse_polynomial_printing) fputc (' ', file);
      fputc (sign < 0 ? '-' : '+', file);
      if (sparse_polynomial_printing) fputc (' ', file);
      print_abs_term (q->term, file);
    }
  }
}

// Building a polynomial from sorted and merged terms.

static Polynomial * build_polynomial () {
  Polynomial * res = 0;
  int i = num_terms;
  while (i > 0)
    res = new_polynomial (terms[--i], res);
  return res;
}

// Here are the most important top level operations on polynomials.

static Polynomial * add_polynomials (Polynomial * a, Polynomial * b) {
  for (Polynomial * p = a; p; p = p->rest) push_term (p->term);
  for (Polynomial * q = b; q; q = q->rest) push_term (q->term);
  sort_terms ();
  merge_terms ();
  Polynomial * res = build_polynomial ();
  clear_terms ();
  return res;
}

static Polynomial * multiply_polynomials (Polynomial * a, Polynomial * b) {
  for (Polynomial * p = a; p; p = p->rest) {
    for (Polynomial * q = b; q; q = q->rest) {
      Term * t = multiply_terms (p->term, q->term);
      push_term (t);
      delete_term (t);
    }
  }
  sort_terms ();
  merge_terms ();
  Polynomial * res = build_polynomial ();
  clear_terms ();
  return res;
}

static int cmp_polymomials (Polynomial * a, Polynomial * b) {
  const Polynomial * p = a;
  const Polynomial * q = b;
  while (p && q) {
    int res = cmp_terms (p->term, q->term);
    if (res) return res;
    p = p->rest;
    q = q->rest;
  }
  if (!p && q) return -1;
  if (p && !q) return 1;
  return 0;
}

static uint64_t hash_polynomial (Polynomial * a) {
  size_t i = 0;
  uint64_t res = 0;
  for (Polynomial * p = a; p; p = p->rest) {
    res += hash_term (p->term);
    res *= nonces[i];
    if (++i == num_nonces) i = 0;
  }
  return res;
}

static uint64_t size_polynomial (Polynomial * a) {
  uint64_t res = 0;
  while (a) res++, a = a->rest;
  return res;
}

static uint64_t degree_polynomial (Polynomial * a) {
  uint64_t res = 0;
  while (a) {
    uint64_t tmp = degree_term (a->term);
    if (tmp > res) res = tmp;
    a = a->rest;
  }
  return res;
}

/*------------------------------------------------------------------------*/

static size_t size_buffer;
static size_t num_buffer;
static char * buffer;

static void enlarge_buffer () {
  size_t new_size_buffer = size_buffer ? 2*size_buffer : 1;
  buffer = reallocate (buffer, size_buffer, new_size_buffer);
  size_buffer = new_size_buffer;
}

static void push_buffer (char ch) {
  if (size_buffer == num_buffer) enlarge_buffer ();
  buffer[num_buffer++] = ch;
}

static void clear_buffer () { num_buffer = 0; }

static void deallocate_buffer () { DEALLOCATE (buffer, size_buffer); }

/*------------------------------------------------------------------------*/

// Character reading routines.

static uint64_t lineno;
static uint64_t columno;
static uint64_t charno;
static const char * parse_file_name;
static FILE * parse_file;
static int close_parse_file;
static int saved_char;
static int char_saved;
static uint64_t last_line_length;

static int next_char () {
  int res;
  if (char_saved) {
    res = saved_char;
    char_saved = 0;
  } else {
#ifdef HAVEUNLOCKEDIO
    res = getc_unlocked (parse_file);
#else
    res = getc (parse_file);
#endif
  }
  if (res == '\n') {
    last_line_length = columno;
    columno = 0;
    lineno++;
  } else if (res != EOF) {
    columno++;
  }
  if (res != EOF) charno++;
  return res;
}

static void prev_char (int ch) {
  assert (!char_saved);
  if (ch == '\n') {
    columno = last_line_length;
    assert (lineno > 0);
    lineno--;
  } else if (ch != EOF) {
    assert (charno > 0);
    charno--;
    assert (columno > 0);
    columno--;
  }
  saved_char = ch;
  char_saved = 1;
}

/*------------------------------------------------------------------------*/

// Scanner.

typedef const char * Token;

static Token END_OF_FILE_TOKEN = "end-of-file";
static Token MINUS_TOKEN       = "minus operator";
static Token PLUS_TOKEN	       = "addition operator";
static Token MULTIPLY_TOKEN    = "multiplication operator";
static Token POWER_TOKEN       = "exponentiation operator";
static Token COLON_TOKEN       = "colon separator";
static Token COMMA_TOKEN       = "comma separator";
static Token SEMICOLON_TOKEN   = "semicolon separator";
static Token NUMBER_TOKEN      = "number";
static Token VARIABLE_TOKEN    = "variable";

static Token token;
static uint64_t lineno_at_start_of_last_token;
static uint64_t columno_at_start_of_last_token;

static void parse_error (const char * msg, ...) {
  fflush (stdout);
  fprintf (stderr,
    "*** 'pactrim' parse error in '%s'"
    " line %" PRIu64 " column %" PRIu64,
    parse_file_name,
    lineno_at_start_of_last_token,
    columno_at_start_of_last_token);
  if (buffer[0] && isprint (buffer[0]))
    fprintf (stderr, " at '%s'", buffer);
  else if (token == END_OF_FILE_TOKEN)
    fputs (" at end-of-file", stderr);
  fputs (": ", stderr);
  va_list ap;
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fputc ('\n', stderr);
  fflush (stderr);
  exit (1);
}

static Token new_token (Token t) {
  push_buffer (0);
  token = t;
#ifdef VERBOSEPARSING
  printf ("%s:%" PRIu64 ":%" PRIu64 ": %-24s %s\n",
    parse_file_name,
    lineno_at_start_of_last_token,
    columno_at_start_of_last_token,
    t, buffer);
#endif
  return token;
}

static Token next_token () {
  clear_buffer ();
  for (;;) {
    int ch = next_char ();
    if (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n')
      continue;
    lineno_at_start_of_last_token = lineno;
    columno_at_start_of_last_token = columno;
    if (ch == EOF) return new_token (END_OF_FILE_TOKEN);
    push_buffer (ch);
    if (ch == '-') return new_token (MINUS_TOKEN);
    if (ch == '+') return new_token (PLUS_TOKEN);
    if (ch == '*') return new_token (MULTIPLY_TOKEN);
    if (ch == '^') return new_token (POWER_TOKEN);
    if (ch == ':') return new_token (COLON_TOKEN);
    if (ch == ',') return new_token (COMMA_TOKEN);
    if (ch == ';') return new_token (SEMICOLON_TOKEN);
    if ('0' <= ch && ch <= '9') {
      while ('0' <= (ch = next_char ()) && ch <= '9')
	push_buffer (ch);
      prev_char (ch);
      return new_token (NUMBER_TOKEN);
    }
    if (is_valid_variable_first_letter (ch)) {
      while (is_valid_variable_letter (ch = next_char ()))
	push_buffer (ch);
      prev_char (ch);
      return new_token (VARIABLE_TOKEN);
    }
    if (isprint (ch)) parse_error ("invalid character");
    else parse_error ("invalid character code 0x%02x", ch);
  }
}

static int is_separator_token (Token t) {
  if (t == COLON_TOKEN) return 1;
  if (t == COMMA_TOKEN) return 1;
  if (t == SEMICOLON_TOKEN) return 1;
  return 0;
}

/*------------------------------------------------------------------------*/

// Parser for polynomials.

static Number * parse_natural () {
  if (buffer[0] == '0' && buffer[1])
    parse_error ("invalid number");
  Number * res = new_number (buffer);
  next_token ();
  return res;
}

static Variable * parse_variable () {
  Variable * res = new_variable (buffer);
  next_token ();
  return res;
}

static uint64_t parse_exponent () {
  uint64_t res = 0;
  for (const char * p = buffer; *p; p++) {
    if (UINT64_MAX/10 < res)
      parse_error ("invalid exponent");
    res *= 10;
    int ch = *p;
    assert ('0' <= ch);
    assert (ch <= '9');
    uint64_t digit = ch - '0';
    if (res - digit < res)
      parse_error ("invalid exponent");
    res += digit;
  }
  next_token ();
  return res;
}

static Monomial * parse_monomial () {
  assert (!num_powers);
  while (token == VARIABLE_TOKEN) {
    Variable * variable = parse_variable ();
    uint64_t exponent;
    if (token == POWER_TOKEN) {
      next_token ();
      if (token != NUMBER_TOKEN)
	parse_error ("exponent missing after '^'");
      exponent = parse_exponent ();
    } else exponent = 1;
    Power p = new_power (variable, exponent);
    push_power (p);
    if (token == MULTIPLY_TOKEN) next_token ();
  }
  sort_powers ();
  merge_powers ();
  Monomial * res = build_monomial ();
  clear_powers ();
  return res;
}

static Term * parse_term (int sign) {
  Number * number = 0;
  if (token == NUMBER_TOKEN) number = parse_natural ();
  else if (token == VARIABLE_TOKEN) number = new_number ("1");
  else parse_error ("expected term");
  if (sign < 0) {
    Number * tmp = negate_number (number);
    delete_number (number);
    number = tmp;
  }
  if (token == MULTIPLY_TOKEN) next_token ();
  Monomial * monomial = parse_monomial ();
  Term * res = new_term (number, monomial);
  delete_monomial (monomial);
  delete_number (number);
  return res;
}

static Polynomial * parse_polynomial () {
  int sign;
  if (token == MINUS_TOKEN) {
    next_token ();
    if (token == NUMBER_TOKEN && buffer[0] == '0')
      parse_error ("unexpected '0' after '-'");
    sign = -1;
  } else sign = 1;
  assert (!num_terms);
  for (;;) {
    Term * term = parse_term (sign);
    push_term (term);
    delete_term (term);
    if (is_separator_token (token)) break;
    if (token == MINUS_TOKEN) {
      sign = -1;
      next_token ();
    } else if (token == PLUS_TOKEN) {
      sign = 1;
      next_token ();
    } else parse_error ("unexpected %s", token);
  }
  sort_terms ();
  merge_terms ();
  Polynomial * res = build_polynomial ();
  clear_terms ();
  return res;
}

/*------------------------------------------------------------------------*/

typedef const char * Rule;

static Rule ORIGINAL_RULE       = "original polynomial";
static Rule ADDITION_RULE       = "polynomial addition";
static Rule MULTIPLICATION_RULE = "polynomial multiplication";

typedef struct Inference Inference;

struct Inference {
  Rule rule;
  uint64_t id;				// unique
  uint64_t core;			// used that often in core
  uint64_t last_used_id;		// id of last use as antecedent
  const char * file;			// file name
  uint64_t line;			// start line in file
  uint64_t column;			// start column in file
  Inference * antecedent[2];		// 1 or 2 existing polynomials
  Polynomial * factor;			// arbitrary multiplicand
  Polynomial * conclusion;		// produced / added polynomial
  Inference * chronological_next;	// next doubly linked list
  Inference * chronological_prev;	// prev doubly linked list
  Inference * collision_chain_link;	// for hash table
  uint64_t hash;			// saved for faster enlargement
};

// inferences hash table with multi-set semantics
// (chains chronologically sorted thus first inference found first)

static uint64_t num_inferences;
static uint64_t size_inferences;
static Inference ** inference_table;

// inferences chronological doubly linked list

static Inference * first_inference;
static Inference * last_inference;

// statistics for inferences

static uint64_t original_inferences;
static uint64_t addition_inferences;
static uint64_t multiplication_inferences;
static uint64_t rule_inferences;
static uint64_t searched_inferences;
static uint64_t collisions_inferences;
static uint64_t non_zero_constant_polynomial_inferences;

static void enlarge_inferences () {
  uint64_t new_size_inferences = size_inferences ? 2*size_inferences : 1;
  Inference ** new_inference_table;
  ALLOCATE (new_inference_table, new_size_inferences);
  for (Inference * p = last_inference; p; p = p->chronological_prev) {
    uint64_t h = p->hash & (new_size_inferences - 1);
    Inference * m = new_inference_table[h];
    p->collision_chain_link = m;
    new_inference_table[h] = p;
  }
  DEALLOCATE (inference_table, size_inferences);
  inference_table = new_inference_table;
  size_inferences = new_size_inferences;
}

static Inference *
allocate_inference (Rule rule,
		    const char * file,
		    uint64_t line,
		    uint64_t column,
		    Inference * antecedent0,
		    Inference * antecedent1,
		    Polynomial * factor,
		    Polynomial * conclusion,
		    uint64_t hash)
{
  Inference * res = allocate (sizeof *res);
  res->rule = rule;
  res->id = ++num_inferences;
  res->core = 0;
  res->file = file;
  res->line = line;
  res->column = column;
  res->antecedent[0] = antecedent0;
  res->antecedent[1] = antecedent1;
  res->factor = factor;
  res->conclusion = conclusion;
  if (!last_inference) first_inference = res;
  else last_inference->chronological_next = res;
  res->chronological_prev = last_inference;
  last_inference = res;
  res->chronological_next = 0;
  res->collision_chain_link = 0;
  res->hash = hash;
  if (rule == ORIGINAL_RULE) original_inferences++;
  if (rule == ADDITION_RULE)
    addition_inferences++, rule_inferences++;
  if (rule == MULTIPLICATION_RULE)
    multiplication_inferences++, rule_inferences++;
  if (is_non_zero_constant_polynomial (conclusion))
    non_zero_constant_polynomial_inferences++;
  return res;
}

// Find first inference with the same polynomial.

static Inference * find_inference (Polynomial * p) {
  if (!size_inferences) return 0;
  const uint64_t hash = hash_polynomial (p);
  const uint64_t h = hash & (size_inferences - 1);
  searched_inferences++;
  Inference * res;
  for (res = inference_table[h];
       res && cmp_polymomials (res->conclusion, p);
       res = res->collision_chain_link)
    collisions_inferences++;
  return res;
}

// Add inference (even if polynomial was already inferred).

static Inference *
new_inference (Rule rule,
               const char * file,
	       uint64_t line,
	       uint64_t column,
	       Inference * antecedent0,
	       Inference * antecedent1,
	       Polynomial * factor,
	       Polynomial * conclusion)
{
  if (num_inferences == size_inferences) enlarge_inferences ();
  const uint64_t hash = hash_polynomial (conclusion);
  const uint64_t h = hash & (size_inferences - 1);
  Inference ** p, * res;
  searched_inferences++;
  for (p = inference_table + h;
       p && (res = *p);
       p = &res->collision_chain_link)
    collisions_inferences++;
  res = allocate_inference (rule, file, line, column,
			    antecedent0, antecedent1, factor,
			    conclusion, hash);
  if (p) *p = res;
  return res;
}

static void delete_inference (Inference * i) {
  delete_polynomial (i->factor);
  delete_polynomial (i->conclusion);
  deallocate (i, sizeof *i);
}

static void delete_inferences () {
  for (Inference * p = first_inference, * n; p; p = n) {
    n = p->chronological_next;
    delete_inference (p);
  }
  DEALLOCATE (inference_table, size_inferences);
#ifdef TESTONLY
  inference_table = 0;
  size_inferences = num_inferences = 0;
  first_inference = last_inference = 0;
  original_inferences = 0;
  addition_inferences = 0;
  multiplication_inferences = 0;
  searched_inferences = 0;
  collisions_inferences = 0;
  non_zero_constant_polynomial_inferences = 0;
#endif
}

/*------------------------------------------------------------------------*/

// Proof statistics.

static uint64_t size_proof;
static uint64_t degree_proof;
static uint64_t length_proof;

/*------------------------------------------------------------------------*/

// Original polynomial inference parsing.

static void init_parsing (const char * file_name) {
  if (file_name) {
    parse_file_name = file_name;
    parse_file = fopen (file_name, "r");
    if (!parse_file)
      die ("can not open '%s' for reading", file_name);
    close_parse_file = 1;
  } else {
    parse_file = stdin;
    parse_file_name = "<stdin>";
    close_parse_file = 0;
  }
  lineno = 1;
  columno = 0;
  charno = 0;
}

static void reset_parsing () {
  if (close_parse_file && fclose (parse_file))
    die ("failed to close '%s'", parse_file_name);
  msg ("read %" PRIu64 " bytes from '%s'", charno, parse_file_name);
}

static void update_statistics_for_newly_added_polynomial (Polynomial * p) {
    const uint64_t size = size_polynomial (p);
    const uint64_t degree = degree_polynomial (p);
    length_proof++;
    size_proof += size;
    if (degree > degree_proof) degree_proof = degree;
#ifdef VERBOSEPARSING
    printf (
      "polynomial %" PRIu64 ", degree %" PRIu64 ", size %" PRIu64 "\n",
      length_proof, degree, size);
    print_polynomial (p, stdout);
    fputc ('\n', stdout);
    fflush (stdout);
#endif
}

static void parse_original_polynomial () {
  uint64_t line = lineno_at_start_of_last_token;
  uint64_t column = columno_at_start_of_last_token;
  Polynomial * p = parse_polynomial ();
  assert (is_separator_token (token));
  update_statistics_for_newly_added_polynomial (p);
  new_inference (ORIGINAL_RULE,
    parse_file_name, line, column, 0, 0, 0, p);
}

/*------------------------------------------------------------------------*/

// Proof rule inference parsing.

static int print_failing_polynomials = 0;

static void
polynomial_not_found (
  Polynomial * p,
  Rule rule,
  uint64_t count,
  const char * which_polynomial,
  uint64_t rule_line,
  uint64_t rule_column,
  uint64_t polynomial_line,
  uint64_t polynomial_column) 
{
  fflush (stdout);
  fprintf (stderr,
    "*** 'pactrim' error in %s rule %" PRIu64 " in '%s'"
    " line %" PRIu64 " column %" PRIu64 ": %s polynomial",
    rule, count, parse_file_name,
    rule_line, rule_column, which_polynomial);
  if (rule_line != polynomial_line)
    fprintf (stderr, " line %" PRIu64, polynomial_line);
  fprintf (stderr, " column %" PRIu64 " not found", polynomial_column);
  if (print_failing_polynomials) {
    fputs (":\n", stderr);
    print_polynomial (p, stderr);
  } else fputs (" (use '--print' to print failing polynomials)", stderr);
  fputc ('\n', stderr);
  fflush (stderr);
  exit (1);
}

static void
polynomials_do_not_match (
  Polynomial * actual,
  Polynomial * expected,
  Rule rule,
  uint64_t count,
  uint64_t rule_line,
  uint64_t rule_column,
  uint64_t polynomial_line,
  uint64_t polynomial_column) 
{
  fflush (stdout);
  fprintf (stderr,
    "*** 'pactrim' error in %s rule %" PRIu64 " in '%s'"
    " line %" PRIu64 " column %" PRIu64 ":"
    " conclusion polynomial",
    rule, count, parse_file_name, rule_line, rule_column);
  if (rule_line != polynomial_line)
    fprintf (stderr, " line %" PRIu64, polynomial_line);
  fprintf (stderr, " column %" PRIu64, polynomial_column);
  if (print_failing_polynomials) {
    fputs (":\n", stderr);
    print_polynomial (actual, stderr);
    fputs ("\ndoes not match expected result:\n", stderr);
    print_polynomial (expected, stderr);
  } else fputs (" does not match expected result"
                " (use '--print' to print failing polynomials)", stderr);
  fputc ('\n', stderr);
  fflush (stderr);
  exit (1);
}

static void parse_practical_algebraic_calculus_rule (uint64_t checked) {
  uint64_t rule_line = lineno_at_start_of_last_token;
  uint64_t rule_column = columno_at_start_of_last_token;
  Rule rule = 0;
  if (token == PLUS_TOKEN) rule = ADDITION_RULE;
  else if (token == MULTIPLY_TOKEN) rule = MULTIPLICATION_RULE;
  else parse_error ("expected '+' or '*' instead of %s", token);
  next_token ();
  if (!is_separator_token (token))
    parse_error ("expected separator instead of %s", token);
  next_token ();
  uint64_t p0_line = lineno_at_start_of_last_token;
  uint64_t p0_column = columno_at_start_of_last_token;
  Polynomial *p0 = parse_polynomial ();
  assert (is_separator_token (token));
  Inference * i0 = find_inference (p0);
  if (!i0)
    polynomial_not_found (p0,
      rule, checked + 1, "first",
      rule_line, rule_column, p0_line, p0_column);
  delete_polynomial (p0);
  next_token ();
  uint64_t p1_line = lineno_at_start_of_last_token;
  uint64_t p1_column = columno_at_start_of_last_token;
  Polynomial *p1 = parse_polynomial ();
  assert (is_separator_token (token));
  Inference * i1;
  if (rule == ADDITION_RULE) {
    i1 = find_inference (p1);
    if (!i1)
      polynomial_not_found (p1,
	rule, checked + 1, "second",
	rule_line, rule_column, p1_line, p1_column);
    delete_polynomial (p1);
    p1 = 0;
  } else i1 = 0;
  next_token ();
  uint64_t p2_line = lineno_at_start_of_last_token;
  uint64_t p2_column = columno_at_start_of_last_token;
  Polynomial *p2 = parse_polynomial ();
  assert (is_separator_token (token));
  Polynomial * conclusion;
  if (rule == ADDITION_RULE) {
    assert (i0), assert (i1), assert (!p1);
    conclusion = add_polynomials (i0->conclusion, i1->conclusion);
  } else {
    assert (i0), assert (!i1); // 'p1' might be empty so zero
    conclusion = multiply_polynomials (i0->conclusion, p1);
  }
  if (cmp_polymomials (p2, conclusion))
    polynomials_do_not_match (p2, conclusion,
      rule, checked + 1, rule_line, rule_column, p2_line, p2_column);
  delete_polynomial (p2);
  update_statistics_for_newly_added_polynomial (conclusion);
  new_inference (rule,
    parse_file_name, rule_line, rule_column, i0, i1, p1, conclusion);
}

/*------------------------------------------------------------------------*/

// Parse whole files.

static void parse_original_polynomials (const char * file_name) {
  init_parsing (file_name);
  msg ("reading original polynomials from '%s'", parse_file_name);
  uint64_t original = 0;
  while (next_token () != END_OF_FILE_TOKEN)
    parse_original_polynomial (), original++;
  msg ("found %" PRIu64 " original polynomials in '%s'",
    original, parse_file_name);
  reset_parsing ();
}

static void
parse_practical_algebraic_calculus_rules (const char * file_name) {
  init_parsing (file_name);
  msg ("reading polynomial algebraic calculus proof from '%s'",
    parse_file_name);
  uint64_t checked = 0, additions = 0, multiplications = 0;
  while (next_token () != END_OF_FILE_TOKEN) {
    if (token == PLUS_TOKEN) additions++;
    if (token == MULTIPLY_TOKEN) multiplications++;
    parse_practical_algebraic_calculus_rule (checked);
    checked++;
  }
  assert (checked == additions + multiplications);
  msg ("found and checked %" PRIu64 " inferences in '%s'",
    checked, parse_file_name);
  reset_parsing ();
}

static void
parse_mixed_polynomials_and_rules (const char * file_name) {
  init_parsing (file_name);
  msg ("WARNING: mixed mode might lead to unexpected false negatives");
  msg ("reading both original polynomials and rules from '%s'",
    parse_file_name);
  uint64_t original = 0, checked = 0, additions = 0, multiplications = 0;
  while (next_token () != END_OF_FILE_TOKEN) {
    if (token == PLUS_TOKEN || token == MULTIPLY_TOKEN) {
      if (token == PLUS_TOKEN) additions++;
      if (token == MULTIPLY_TOKEN) multiplications++;
      parse_practical_algebraic_calculus_rule (checked);
      checked++;
    } else parse_original_polynomial (), original++;
  }
  assert (checked == additions + multiplications);
  msg ("found %" PRIu64 " original polynomials in '%s'",
    original, parse_file_name);
  msg ("found and checked %" PRIu64 " inferences in '%s'",
    checked, parse_file_name);
  reset_parsing ();
}

/*------------------------------------------------------------------------*/

static double percent (double a, double b) { return b ? 100.0*a/b : 0; }
static double average (double a, double b) { return b ? a/b : 0; }

/*------------------------------------------------------------------------*/

// Core handling.

static const char * core_file_name;
static const char * lemma_file_name;
static const char * unused_file_name;
static const char * redundant_file_name;

static void write_core () {
  assert (core_file_name);
  FILE * file = fopen (core_file_name, "w");
  uint64_t count = 0;
  if (!file) die ("can not write core to '%s'", core_file_name);
  for (Inference * i = first_inference; i; i = i->chronological_next) {
    if (!i->core) continue;
    if (i->rule != ORIGINAL_RULE) continue;
    print_polynomial (i->conclusion, file);
    fputs (";\n", file);
    count++;
  }
  if (fclose (file)) die ("failed to close '%s' core", core_file_name);
  msg ("wrote %" PRIu64 " core polynomials to '%s'", count, core_file_name);
}

static void write_lemmas () {
  assert (lemma_file_name);
  FILE * file = fopen (lemma_file_name, "w");
  uint64_t count = 0;
  if (!file) die ("can not write lemmas to '%s'", lemma_file_name);
  for (Inference * i = first_inference; i; i = i->chronological_next) {
    if (!i->core) continue;
    if (i->rule == ADDITION_RULE) fputc ('+', file);
    else if (i->rule == MULTIPLICATION_RULE) fputc ('*', file);
    else continue;
    fputc (':', file);
    print_polynomial (i->antecedent[0]->conclusion, file);
    if (i->antecedent[1]) {
      fputc (',', file);
      print_polynomial (i->antecedent[1]->conclusion, file);
    }
    if (i->factor) {
      fputc (',', file);
      print_polynomial (i->factor, file);
    }
    fputc (',', file);
    print_polynomial (i->conclusion, file);
    fputs (";\n", file);
    count++;
  }
  if (fclose (file)) die ("failed to close '%s' lemmas", lemma_file_name);
  msg ("wrote %" PRIu64 " lemmas to '%s'", count, lemma_file_name);
}

static void write_unused () {
  assert (unused_file_name);
  FILE * file = fopen (unused_file_name, "w");
  uint64_t count = 0;
  if (!file)
    die ("can not write unused original to '%s'", unused_file_name);
  for (Inference * i = first_inference; i; i = i->chronological_next) {
    if (i->core) continue;
    if (i->rule != ORIGINAL_RULE) continue;
    print_polynomial (i->conclusion, file);
    fputs (";\n", file);
    count++;
  }
  if (fclose (file))
    die ("failed to close '%s' unused original", unused_file_name);
  msg ("wrote %" PRIu64 " unused polynomials to '%s'",
    count, unused_file_name);
}

static void write_redundant () {
  assert (redundant_file_name);
  FILE * file = fopen (redundant_file_name, "w");
  uint64_t count = 0;
  if (!file)
    die ("can not write redundant lemmas to '%s'", redundant_file_name);
  for (Inference * i = first_inference; i; i = i->chronological_next) {
    if (i->core) continue;
    if (i->rule == ADDITION_RULE) fputc ('+', file);
    else if (i->rule == MULTIPLICATION_RULE) fputc ('*', file);
    else continue;
    fputc (':', file);
    print_polynomial (i->antecedent[0]->conclusion, file);
    if (i->antecedent[1]) {
      fputc (',', file);
      print_polynomial (i->antecedent[1]->conclusion, file);
    }
    if (i->factor) {
      fputc (',', file);
      print_polynomial (i->factor, file);
    }
    fputc (',', file);
    print_polynomial (i->conclusion, file);
    fputs (";\n", file);
    count++;
  }
  if (fclose (file))
    die ("failed to close '%s' redundant lemmas", redundant_file_name);
  msg ("wrote %" PRIu64 " redundant lemmas to '%s'",
    count, redundant_file_name);
}

static void compute_core () {
  assert (non_zero_constant_polynomial_inferences > 0);
  Inference * i = first_inference;
  while (assert (i), !is_non_zero_constant_polynomial (i->conclusion))
    i = i->chronological_next;
  msg ("computing core of %s non-zero constant polynomial %" PRIu64,
    (non_zero_constant_polynomial_inferences == 1 ?  "single" : "first"),
    i->id);
  uint64_t core_length = 0, core_size = 0;
  uint64_t original_core = 0, lemma_core = 0;
  uint64_t unused_original = 0, redundant_lemmas = 0;
  for (i->core = 1; i; i = i->chronological_prev) {
    if (i->core) {
      if (i->rule == ORIGINAL_RULE) original_core++; else lemma_core++;
      if (i->antecedent[0]) i->antecedent[0]->core++;
      if (i->antecedent[1]) i->antecedent[1]->core++;
      core_size += size_polynomial (i->conclusion);
      core_length++;
    } else if (i->rule == ORIGINAL_RULE) unused_original++;
    else redundant_lemmas++;
  }
  msg ("original core of %" PRIu64 " polynomials (%.0f%% of core length)",
    original_core, percent (original_core, core_length));
  if (core_file_name) write_core ();
  msg ("lemma core of %" PRIu64 " polynomials (%.0f%% of core length)",
    lemma_core, percent (lemma_core, core_length));
  if (lemma_file_name) write_lemmas ();
  msg ("core length of %" PRIu64 " polynomials (%.0f%% of proof length)",
    core_length, percent (core_length, num_inferences));
  msg ("core size of %" PRIu64 " monomials (%.0f%% of proof size)",
    core_size, percent (core_size, size_proof));
  msg ("unused %" PRIu64 " original polynomials (%.0f%%)",
    unused_original, percent (unused_original, original_inferences));
  if (unused_file_name) write_unused ();
  msg ("redundant %" PRIu64 " proof lemmas (%.0f%%)",
    redundant_lemmas, percent (redundant_lemmas, rule_inferences));
  if (redundant_file_name) write_redundant ();

}

/*------------------------------------------------------------------------*/

static void init () {
  init_nonces ();
}

static void deallocate_all () {
  delete_inferences ();
  deallocate_monomials ();
  deallocate_terms ();
  deallocate_powers ();
  deallocate_variables ();
  deallocate_buffer ();
  assert (!current_allocated_bytes);
}

static void reset () {
  deallocate_all ();
}

/*------------------------------------------------------------------------*/

#ifdef HAVEGETRUSAGE
#include <sys/time.h>
#include <sys/resource.h>
#endif

static void print_statistics () {
  msg ("proof length %" PRIu64 " (number of polynomials)", length_proof);
  msg ("proof size %" PRIu64 " (on average %.1f monomials per polynomial)",
    size_proof, average (size_proof, length_proof));
  msg ("proof degree %" PRIu64 " (internal maximum degree %" PRIu64 ")",
    degree_proof, max_degree);
  msg ("searched %" PRIu64 " inferences %.1f average collisions",
    searched_inferences,
    average (collisions_inferences, searched_inferences));
  msg ("%" PRIu64 " inferences, %.1f average searches",
    num_inferences, average (searched_inferences, num_inferences));
  msg ("original inferences %" PRIu64 " (%.0f%%)",
    original_inferences, percent (original_inferences, num_inferences));
  msg ("addition inferences %" PRIu64 " (%.0f%%)",
    addition_inferences, percent (addition_inferences, num_inferences));
  msg ("multiplication inferences %" PRIu64 " (%.0f%%)",
    multiplication_inferences,
    percent (multiplication_inferences, num_inferences));
  msg ("maximum %" PRIu64 " of total %" PRIu64 " monomials (%.0f%%)",
    max_monomials, total_monomials,
    percent (max_monomials, total_monomials));
  msg ("searched %" PRIu64 " monomials %.0f%% hits %.1f average collisions",
    searched_monomials,
    percent (hits_monomials, searched_monomials),
    average (collisions_monomials, searched_monomials));
  msg ("maximum %" PRIu64 " of total %" PRIu64 " numbers (%.0f%%)",
    max_numbers, total_numbers, percent (max_numbers, total_numbers));
  msg ("maximum %zu bytes allocated (%.1f MB)",
    max_allocated_bytes, max_allocated_bytes/(double)(1<<20));
#ifdef HAVEGETRUSAGE
  struct rusage u;
  if (getrusage (RUSAGE_SELF, &u)) return;
  size_t s = ((size_t) u.ru_maxrss) << 10;
  msg ("maximum resident set size %zu bytes (%.1f MB)",
    s, s/(double)(1<<20));
  double t = u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
  t += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
  msg ("process time %.3f seconds", t);
#endif
}

/*------------------------------------------------------------------------*/
#ifdef TESTONLY
/*------------------------------------------------------------------------*/

// Internal test code for development and debugging purposes only.

static int pick (int l, int r) {
  assert (l <= r);
  return l + (rand () % (r - l + 1));
}

static char pick_char () { return pick ('a', 'z'); }

static char pick_digit () { return pick ('0', '9'); }

static Number * pick_number_aux (int minimum_sign, int max_len) {
  int l = pick (0, max_len);
  int sign = pick (0, 1) ? -1 : 1;
  if (sign < minimum_sign) sign = minimum_sign;
  if (sign < 0) push_buffer ('-');
  for (int j = 0; j < l; j ++) push_buffer (pick_digit ());
  push_buffer (0);
  Number * res = new_number (buffer);
  clear_buffer ();
  return res;
}

static Number * pick_number (int n) { return pick_number_aux (-1, n); }
static Number * pick_natural (int n) { return pick_number_aux (1, n); }

static void println_variable (Variable * v) {
  print_variable (v, stdout);
  fputc ('\n', stdout);
}

static void println_number (Number * n) {
  print_number (n, stdout);
  fputc ('\n', stdout);
}

static void test_variables () {
  printf ("test_variables\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    int l = pick (5, 20);
    for (int j = 0; j < l; j ++)
      push_buffer (pick_char ());
    push_buffer (0);
    Variable * v = new_variable (buffer);
    clear_buffer ();
    println_variable (v);
  }
}

static void test_hash_numbers () {
  printf ("test_hash_numbers\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    Number * a = pick_number (60);
    printf ("0x%016llx ", (unsigned long long) hash_number (a));
    println_number (a);
    delete_number (a);
  }
}

static void test_add_numbers () {
  printf ("test_add_numbers\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    Number * a = pick_number (10);
    Number * b = pick_number (10);
    Number * c = add_numbers (a, b);
    print_number (a, stdout);
    fputs (" + ", stdout);
    print_number (b, stdout);
    fputs (" = ", stdout);
    println_number (c);
    assert (atol (a) + atol (b) == atol (c));
    delete_number (a);
    delete_number (b);
    delete_number (c);
  }
}

static void test_multiply_natural_step () {
  printf ("test_multiply_natural_step\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    Number * a = pick_natural (15);
    char d = pick_digit ();
    size_t e = pick (0, 3);
    Number * b = multiply_natural_step (a, d, e);
    print_number (a, stdout);
    printf (" * %c * 10^%zu = ", d, e);
    println_number (b);
    long ee = 1;
    for (size_t i = 0; i < e; i++) ee *= 10;
    assert (atol (a) * (d - '0') * ee == atol (b));
    delete_number (a);
    delete_number (b);
  }
}

static void test_multiply_numbers () {
  printf ("test_multiply_numbers\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    Number * a = pick_number (8);
    Number * b = pick_number (8);
    Number * c = multiply_numbers (a, b);
    print_number (a, stdout);
    fputs (" * ", stdout);
    print_number (b, stdout);
    fputs (" = ", stdout);
    println_number (c);
    fflush (stdout);
    assert (atol (a) * atol (b) == atol (c));
    delete_number (a);
    delete_number (b);
    delete_number (c);
  }
}

static void test_numbers () {
  test_hash_numbers ();
  test_add_numbers ();
  test_multiply_natural_step ();
  test_multiply_numbers ();
}

static Power pick_power (char l, char r, int e) {
  assert ('a' <= l);
  assert (l <= r);
  assert (r <= 'z');
  assert (0 <= e);
  char name[2];
  name[0] = pick (l, r);
  name[1] = 0;
  return new_power (new_variable (name), pick (0, e));
}

static void print_powers (FILE * file) {
  if (num_powers) {
    for (size_t i = 0; i < num_powers; i++) {
      if (i) fputc ('*', file);
      print_power (powers[i], file);
    }
  } else fputc ('1', file);
}

static void println_powers () {
  print_powers (stdout);
  fputc ('\n', stdout);
}

static void test_powers () {
  printf ("test_powers\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    int n = pick (0, 25);
    int l = pick ('a', 'z');
    int r = MIN (l + n, 'z');
    int k = pick (1, 10);
    for (int j = 0; j < k; j++)
      push_power (pick_power (l, r, 3));
    print_powers (stdout);
    sort_powers ();
    fputs (" = ", stdout);
    print_powers (stdout);
    merge_powers ();
    fputs (" = ", stdout);
    println_powers ();
    clear_powers ();
  }
}

static void println_monomial (Monomial * m) {
  print_monomial (m, stdout);
  fputc ('\n', stdout);
}

static void test_print_monomials1 () {
  printf ("test_print_monomials1\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    int n = pick (1, 10), prev = 1000;
    Monomial * m = 0;
    for (int j = 0; j < n; j++) {
      prev = prev - pick (1, 50);
      char name[20];
      sprintf (name, "x%3d", prev);
      Variable * v = new_variable (name);
      unsigned exponent = pick (1, 16);
      Monomial * tmp = new_monomial (v,  exponent, m);
      delete_monomial (m);
      m = tmp;
    }
    println_monomial (m);
    delete_monomial (m);
  }
  assert (!current_monomials);
}

static int variable_name_size = 1;
static int variable_range = 3;

static Monomial * pick_monomial () {
  int p = pick (0, 5);
  int l = pick ('a', 'a' + variable_range);
  int r = pick (l, 'a' + variable_range);
  for (int i = 0; i < p; i++) {
    int e = pick (1, 3);
    char name[variable_name_size + 1];
    int n = pick (1, sizeof (name) - 1);
    for (int j = 0; j < n; j++)
      name[j] = pick (l, r);
    name[n] = 0;
    Variable * v = new_variable (name);
    push_power (new_power (v, e));
  }
  sort_powers ();
  merge_powers ();
  Monomial * res = build_monomial ();
  clear_powers ();
  return res;
}

static void test_multiply_monomials1 () {
  printf ("test_multiply_monomials1\n");
  srand (42);
  Variable * x = new_variable ("x");
  Variable * y = new_variable ("y");
  Variable * z = new_variable ("z");
  for (int i = 0; i < 1000; i++) {
    Monomial * a = new_monomial (z, pick (0, 10), 0);
    Monomial * b = new_monomial (y, pick (0, 10), a);
    delete_monomial (a);
    Monomial * c = new_monomial (x, pick (0, 10), b);
    delete_monomial (b);
    Monomial * d = new_monomial (z, pick (0, 10), 0);
    Monomial * e = new_monomial (y, pick (0, 10), d);
    delete_monomial (d);
    Monomial * f = new_monomial (x, pick (0, 10), e);
    delete_monomial (e);
    Monomial * m = multiply_monomials (c, f);
    print_monomial (c, stdout);
    fputs (" * ", stdout);
    print_monomial (f, stdout);
    fputs (" = ", stdout);
    println_monomial (m);
    delete_monomial (c);
    delete_monomial (f);
    delete_monomial (m);
  }
  assert (!current_monomials);
}

static void test_print_monomials2 () {
  printf ("test_print_monomials2\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    Monomial * m = pick_monomial ();
    println_monomial (m);
    delete_monomial (m);
  }
}

static void test_multiply_monomials2 () {
  printf ("test_multiply_monomials2\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    Monomial * a = pick_monomial ();
    Monomial * b = pick_monomial ();
    Monomial * c = multiply_monomials (a, b);
    print_monomial (a, stdout);
    fputs (" * ", stdout);
    print_monomial (b, stdout);
    fputs (" = ", stdout);
    println_monomial (c);
    delete_monomial (a);
    delete_monomial (b);
    delete_monomial (c);
  }
}

static void test_monomials () {
  test_print_monomials1 ();
  test_multiply_monomials1 ();
  test_print_monomials2 ();
  test_multiply_monomials2 ();
}

static Term * pick_term () {
  Number * number = pick_number (3);
  Monomial * monomial = pick_monomial ();
  Term * res = new_term (number, monomial);
  delete_number (number);
  delete_monomial (monomial);
  return res;
}

static void println_term (Term * t) {
  print_term (t, stdout);
  fputc ('\n', stdout);
}

static void test_print_terms () {
  printf ("test_print_terms\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    Term * a = pick_term ();
    println_term (a);
    delete_term (a);
  }
}

static void test_add_terms () {
  printf ("test_add_terms\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    Term * a = pick_term ();
    Number * number = pick_number (9);
    Term * b = new_term (number, a->monomial);
    delete_number (number);
    Term * c = add_terms (a, b);
    print_term (a, stdout);
    fputs (" + ", stdout);
    print_term (b, stdout);
    fputs (" = ", stdout);
    println_term (c);
    delete_term (a);
    delete_term (b);
    delete_term (c);
  }
}

static void test_multiply_terms () {
  printf ("test_multiply_terms\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    Term * a = pick_term ();
    Term * b = pick_term ();
    Term * c = multiply_terms (a, b);
    print_term (a, stdout);
    fputs (" * ", stdout);
    print_term (b, stdout);
    fputs (" = ", stdout);
    println_term (c);
    delete_term (a);
    delete_term (b);
    delete_term (c);
  }
}

static void test_terms () {
  test_print_terms ();
  test_add_terms ();
  test_multiply_terms ();
}

static void print_terms (FILE * file) {
  if (num_terms) {
    for (size_t i = 0; i < num_terms; i++) {
      if (i) fputs (" + ", file);
      print_term (terms[i], file);
    }
  } else fputc ('0', file);
}

void println_terms () {
  print_terms (stdout);
  fputc ('\n', stdout);
}

static void println_polynomial (Polynomial * p) {
  print_polynomial (p, stdout);
  fputc ('\n', stdout);
}

static Polynomial * pick_polynomial () {
  int n = pick (0, 4);
  if (!n && pick (0, 10)) n = 1;
  for (int i = 0; i < n; i++) {
    Term * t = pick_term ();
    push_term (t);
    delete_term (t);
  }
  sort_terms ();
  merge_terms ();
  Polynomial * res = build_polynomial ();
  clear_terms ();
  return res;
}

static void test_print_polynomials () {
  printf ("test_print_polynomials\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    Polynomial * p = pick_polynomial ();
    println_polynomial (p);
    delete_polynomial (p);
  }
}

static void test_add_polynomials () {
  printf ("test_add_polynomials\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    Polynomial * a = pick_polynomial ();
    Polynomial * b = pick_polynomial ();
    Polynomial * c = add_polynomials (a, b);
    fputc ('(', stdout);
    print_polynomial (a, stdout);
    fputs (") + (", stdout);
    print_polynomial (b, stdout);
    fputs (") = ", stdout);
    println_polynomial (c);
    delete_polynomial (a);
    delete_polynomial (b);
    delete_polynomial (c);
  }
}

static void test_multiply_polynomials () {
  printf ("test_multiply_polynomials\n");
  srand (42);
  for (int i = 0; i < 1000; i++) {
    Polynomial * a = pick_polynomial ();
    Polynomial * b = pick_polynomial ();
    Polynomial * c = multiply_polynomials (a, b);
    fputc ('(', stdout);
    print_polynomial (a, stdout);
    fputs (") * (", stdout);
    print_polynomial (b, stdout);
    fputs (") = ", stdout);
    println_polynomial (c);
    delete_polynomial (a);
    delete_polynomial (b);
    delete_polynomial (c);
  }
}

static void test_polynomials () {
  test_print_polynomials ();
  test_add_polynomials ();
  test_multiply_polynomials ();
}

static void test_tokens_example (const char * base) {
  size_t baselen = strlen (base);
  char path[baselen + 20];

  sprintf (path, "examples/%s.polys", base);
  srand (42);
  printf ("test_tokens on '%s'\n", path);
  lineno = 1;
  parse_file_name = path;
  parse_file = fopen (path, "r");
  if (!parse_file) die ("can not read '%s'", path);
  while (next_token () != END_OF_FILE_TOKEN)
    ;
  fclose (parse_file);
  parse_file_name = 0;

  sprintf (path, "examples/%s.pac", base);
  srand (42);
  printf ("test_tokens on '%s'\n", path);
  lineno = 1;
  parse_file_name = path;
  parse_file = fopen (path, "r");
  if (!parse_file) die ("can not read '%s'", path);
  while (next_token () != END_OF_FILE_TOKEN)
    ;
  fclose (parse_file);
  parse_file_name = 0;
}

static void test_tokens () {
  test_tokens_example ("btor4");
}

static void
test_parse_example (const char * base) {
  char path[strlen (base) + 20];
  sprintf (path, "examples/%s.polys", base);
  printf ("test_parse_example on '%s'\n", path);
  parse_original_polynomials (path);
  sprintf (path, "examples/%s.pac", base);
  print_failing_polynomials = 1;
  printf ("test_parse_example on '%s'\n", path);
  parse_practical_algebraic_calculus_rules (path);
}

static void test_parser () {
  test_parse_example ("test1");
  test_parse_example ("btor4");
}

int main () {

  (void) usage;
  (void) msg;

  init ();

  test_variables ();
  test_numbers ();
  test_powers ();
  test_monomials ();
  test_terms ();
  test_polynomials ();
  test_tokens ();
  test_parser ();

  reset ();

  print_statistics ();

  return 0;
}

/*------------------------------------------------------------------------*/
#else // of '#ifdef TESTONLY'
/*------------------------------------------------------------------------*/

// Main routine for stand alone checker.

static void banner () {
  msg ("PACTRIM Version " VERSION);
  msg ("Practical Algebraic Calculus Proof Checker and Trimmer");
  msg ("Copyright (C) 2018, Armin Biere, Johannes Kepler University Linz");
}

int main (int argc, char ** argv) {

  int check_refutation = 1;
  const char * first_file_name = 0;
  const char * second_file_name = 0;

  for (int i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-h") ||
        !strcmp (argv[i], "--help"))
      fputs (usage, stdout), exit (0);
    else if (!strcmp (argv[i], "-S"))
      check_refutation = 0;
    else if (!strcmp (argv[i], "-c")) {
      if (++i == argc) die ("argument to '-c' missing (try '-h')");
      if (core_file_name)
	die ("two core files '%s' and '%s' (try '-h')",
	  core_file_name, argv[i]);
      core_file_name = argv[i];
    } else if (!strcmp (argv[i], "-l")) {
      if (++i == argc) die ("argument to '-l' missing (try '-h')");
      if (lemma_file_name)
	die ("two lemma files '%s' and '%s' (try '-h')",
	  lemma_file_name, argv[i]);
      lemma_file_name = argv[i];
    } else if (!strcmp (argv[i], "-u")) {
      if (++i == argc) die ("argument to '-u' missing (try '-h')");
      if (unused_file_name)
	die ("two unused polynomials files '%s' and '%s' (try '-h')",
	  unused_file_name, argv[i]);
      unused_file_name = argv[i];
    } else if (!strcmp (argv[i], "-r")) {
      if (++i == argc) die ("argument to '-r' missing (try '-h')");
      if (redundant_file_name)
	die ("two redundant lemma files '%s' and '%s' (try '-h')",
	  redundant_file_name, argv[i]);
      redundant_file_name = argv[i];
    } else if (!strcmp (argv[i], "--print"))
      print_failing_polynomials = 1;
    else if (!strcmp (argv[i], "--sparse"))
      sparse_polynomial_printing = 1;
    else if (argv[i][0] == '-')
      die ("invalid command line option '%s' (try '-h')", argv[i]);
    else if (second_file_name)
      die ("too many command line arguments (try '-h')");
    else if (first_file_name) second_file_name = argv[i];
    else first_file_name = argv[i];
  }

  banner ();
  init ();

  if (second_file_name) {
    msg ("clean mode (both polynomials and proof specified)");
    parse_original_polynomials (first_file_name);
    parse_practical_algebraic_calculus_rules (second_file_name);
  } else {
    if (first_file_name) msg ("mixed mode (only mixed file specified)");
    else msg ("mixed mode (no file specified using '<stdin>')");
    parse_mixed_polynomials_and_rules (first_file_name);
  }

  msg ("found %" PRIu64 " non-zero constant polynomial inference%s",
    non_zero_constant_polynomial_inferences,
    non_zero_constant_polynomial_inferences == 1 ? "" : "s");
  if (check_refutation && !non_zero_constant_polynomial_inferences)
    die ("no refutation: inferences correct"
         " but no non-zero constant polynomial inferred");

  if (non_zero_constant_polynomial_inferences)
    compute_core ();

  reset ();
  print_statistics ();

  if (non_zero_constant_polynomial_inferences) msg ("REFUTATION CHECKED");
  else msg ("INFERENCES CHECKED");

  return 0;
}

/*------------------------------------------------------------------------*/
#endif // of '#else' for '#ifdef TESTONLY'
/*------------------------------------------------------------------------*/

